#include "sut.h"

/****************************************************************************
  Manage the range specification
 ****************************************************************************/


/*
   Make a new Argv array containing manager and old args 
   
   One complication is when generating args to be pased to programs
   that invoke a shell, for example, rsh.  In this case, we 
   may need to quote the arguments.
 */
int SUT_Get_exec_args( hostnum, ExeNameF, ExeCtx, ArgsF, ArgsC, Environment,
	        Argv, quote_args )
int hostnum;
char      *(*ExeNameF)();
void      *ExeCtx;
char      **(*ArgsF)();
void      *ArgsC;
char      *Environment;
char      ***Argv;
int       quote_args;
{
char **user_argv, **argv;
int  user_argc, i;

/* Get the arguments for the users program */
if (ArgsF)
    user_argv = (*ArgsF)( hostnum, ArgsC );
else 
    user_argv = (char **)ArgsC;
user_argc = 0;
while (user_argv[user_argc]) user_argc++;

*Argv = argv = (char **) MALLOC( (user_argc + 6 ) * sizeof(char *) );
if (!argv)
    return SUT_NO_MEMORY;

argv[0] = SUT_REMSH;
argv[1] = 0;         /* Hostname location */
argv[2] = ExeNameF ? (*ExeNameF)( hostnum, ExeCtx ) : (char *)ExeCtx;
argv[3] = 0;         /* Hostlist location */
for (i=0; i<user_argc; i++) {
    if (quote_args) {
	argv[4+i] = (char *)MALLOC( strlen(user_argv[i]) + 3 );
	if (!argv[4+i]) return SUT_NO_MEMORY;
	argv[4+i][0] = '\'';
	strcpy( argv[4+i]+1, user_argv[i] );
	strcat( argv[4+i], "'" );
	}
    else
	argv[4+i] = user_argv[i];
    }
argv[4+user_argc] = 0;

return SUT_SUCCESS;
}


char *SUT_Condense( list, count )
int *list, count;
{
static char buf[1024];
char tmp[64];
int  s, e, i, j;

buf[0] = 0;
i = 0;
while (i < count) {
    s = list[i];
    j = 1;
    while (i+j < count && list[i+j] == s + j) j++;
    /* list[i+j] is first element of NEXT range */
    e = list[i+j-1];
    if (j > 1)
	sprintf( tmp, "%d-%d,", s, e );
    else
	sprintf( tmp, "%d,", s );
    strcat( buf, tmp );
    i += j;
    }
/* Remove the trailing comma */
if (buf[0]) 
    buf[strlen(buf)-1] = 0;
return buf;
}

/****************************************************************************
  Deal with failed connections
 ****************************************************************************/

void SUT_Kill_child(proc_id, kill_how, node)
int proc_id;
int kill_how;
char *node;
{
fprintf( stderr, "Timeout or bad handshake - %s - !  Trying next...\n", node);
kill(proc_id, kill_how);
}

/* Handle a local process exit by saving the exit status */
int SUT_Leaf_exit( jobs, i )
SUT_Jobs *jobs;
int      i;
{
SUT_Connection *conn = jobs->conn + i;
int            err;
int            prog_stat;

/* fprintf( stderr, "Leaf exit...\n" ); */
if (!conn->pid) return;

FD_CLR( conn->readfd, &jobs->rdset );
close( conn->readfd );
close( conn->writefd );
conn->readfd   = -1;
conn->writefd  = -1;
conn->state    = SUT_DEAD;
if (conn->pid) {
    errno = EINTR;
    err   = -1;
    do { 
	err = waitpid(conn->pid, &prog_stat, 0 );
	} while (err == -1 && errno == EINTR);
    if (SUT_Debug & SUT_EXITOPS) {
	fprintf( stderr, "Waited for child %d (%d) in leaf exit\n", i, 
		conn->pid );
	}
    if (err == -1 && errno == ECHILD) {
	conn->pid = 0;
	return SUT_SUCCESS;
	}
    if (err != conn->pid && err != 0) 
	perror("Wait failed in SUT_Leaf_exit");
    if (err == conn->pid && err != 0) {
	if (WIFEXITED(prog_stat)) {
	    jobs->retstat |= WEXITSTATUS(prog_stat);
	    conn->stat     = WEXITSTATUS(prog_stat);
	    jobs->nactive --;
	    if (SUT_Debug & SUT_EXITOPS)
		fprintf( stderr, "Decrementing active connections to %d\n",
			 jobs->nactive );
	    }
	}
    conn->pid      = 0;
    }

return SUT_SUCCESS;
}

/* Remove the i'th entry from the jobs array (closing descriptors
   as required).
   In the case that this was a final child, we need to propagate the
   exit status to the parent; otherwise, we just check it (and perhaps
   propagate it as the exit status of the manager) 
 */
int SUT_Do_job_exit( i, jobs )
int i;
SUT_Jobs *jobs;
{
SUT_Connection *conn = jobs->conn + i;
int prog_stat;

if (conn->handle_eof)
    (*conn->handle_eof)( jobs, i );
else {
    FD_CLR( conn->readfd, &jobs->rdset );
    close( conn->readfd );
    close( conn->writefd );
    conn->readfd  = -1;
    conn->writefd = -1;
    if (conn->pid) {
	waitpid(conn->pid, &prog_stat, WNOHANG );
	if (SUT_Debug & SUT_EXITOPS) {
	    fprintf( stderr, "Waited for child %d (%d) in Do jobs exit\n", 
		     i, conn->pid );
	    }
	if (WIFEXITED(prog_stat))
	    conn->stat |= WEXITSTATUS(prog_stat);
	jobs->nactive --;
	if (SUT_Debug & SUT_EXITOPS)
	    fprintf( stderr, "Decrementing active connections to %d\n",
		     jobs->nactive );
	}
    conn->pid = 0;
    }
}


/* Tell our parent that we want to exit and wait for the ok */
int SUT_Exit( jobs )
SUT_Jobs *jobs;
{
SUT_Header buf;

buf.op  = SUT_DONE;
buf.len = jobs->retstat;
WRITE( 1, &buf, sizeof(SUT_Header) );
while (1) {
    READ( 0, &buf, sizeof(SUT_Header) );
    if (buf.op == SUT_ALLOW) continue;
    if (buf.op != SUT_EXIT) 
	fprintf( stderr, "Received unexpected message %d\n", buf.op );
    break;
    }
return SUT_SUCCESS;
}

/****************************************************************************
  Initialization
 ****************************************************************************/

int SUT_Init( jobs )
SUT_Jobs *jobs;
{
int i;

jobs->nactive     = 0;
jobs->nconn       = 0;
jobs->nchildren   = 0;
jobs->maxfds	  = -1;
jobs->availout	  = -1;
jobs->retstat     = 0;
jobs->writeparent = SUT_Writedata;
jobs->readparent  = 0; /* SUT_Readparent */
jobs->writefilter = SUT_Write;
jobs->readin      = 0;
jobs->availout    = SUT_MAX_BUF;
FD_ZERO(&jobs->rdset);

/* Initialize the connection structure */
for (i=0; i<SUT_MAX_LOGP+2; i++) {
    jobs->conn[i].readfd      = -1;
    jobs->conn[i].writefd     = -1;
    jobs->conn[i].pid	      = 0;
    jobs->conn[i].stat	      = 0;
    jobs->conn[i].readdata    = 0;
    jobs->conn[i].writedata   = 0;
    jobs->conn[i].handle_eof  = 0;
    jobs->conn[i].nodenum     = -1;
    jobs->conn[i].state       = SUT_UNOPEN;
    jobs->conn[i].nchild      = 0;
    jobs->conn[i].nlist       = 0;

    jobs->conn[i].linebuf     = (char *)MALLOC( jobs->availout );
    if (!jobs->conn[i].linebuf) return SUT_NO_MEMORY;
    jobs->conn[i].linepos     = 0;
    jobs->conn[i].linelen     = jobs->availout;
    jobs->conn[i].saw_newline = 1;
    jobs->conn[i].ops	      = 0;
    }
jobs->buflen    = 0;
jobs->mynodenum = -1;    
return SUT_SUCCESS;
}

/* Set SIGCHLD handler */
static SUT_Jobs *myjobs = 0;
RETSIGTYPE SUT_handle_child( sig, code, scp )
int               sig, code;
struct sigcontext *scp;
/* Some systems have an additional char *addr */
{
int prog_stat, pid;
int i, j;
SUT_Connection *conn;

/* Really need to block further signals until done ... */
/* fprintf( stderr, "Got SIGCHLD...\n" ); */
pid	   = waitpid( (pid_t)(-1), &prog_stat, WNOHANG );
if (myjobs && pid && (WIFEXITED(prog_stat) || WIFSIGNALED(prog_stat))) {
    if (SUT_Debug & SUT_EXITOPS) {
	fprintf( stderr, "Got signal for child %d (exited)... \n", pid ); 
	}
    /* The child has stopped. Remove it from the jobs array */
    for (i = 0; i<myjobs->nconn; i++) {
	conn = myjobs->conn + i;
	if (conn->pid == pid) {
	    conn->pid  = 0;
	    if (WIFEXITED(prog_stat)) {
		conn->stat |= WEXITSTATUS(prog_stat);
		if (i == myjobs->local) 
		    myjobs->retstat |= conn->stat;
		myjobs->nactive--;
		if (SUT_Debug & SUT_EXITOPS)
		    fprintf( stderr, "Decrementing active connections to %d\n",
			     myjobs->nactive );
		}
	    FD_CLR( conn->readfd, &myjobs->rdset );
	    close( conn->readfd );
	    close( conn->writefd );
	    conn->readfd  = -1;
	    conn->writefd = -1;
	    conn->state = SUT_DEAD;
	    break;
	    }
	}
    /* Child may already have been deleted by Leaf exit; we should
       use conn->state to record it rather than zeroing it */
    /* 
    if (i == myjobs->nconn) {
	fprintf( stderr, "Received signal from unknown child!\n" );
	}
	 */
    }
/* Re-enable signals */
}

/* Note: the AIX /usr/include/sys/signal.h defines the prototype for
   signal as extern void (*signal(int, void (*)(int)))(int), 
   unfortunately, this is wrong (the man page for signal clearly
   shows that the handler is void (*)(int, int, struct sigcontext)
   There isn't anything that we can do about this (other than put the
   signal handler in a file where it doesn't belong....)
 */
int SUT_Set_jobs_data( jobs )
SUT_Jobs *jobs;
{
myjobs = jobs;
#ifdef SUT_SIGCHLD_HANDLER
(void) signal( SIGCHLD, SUT_handle_child );
#endif
return SUT_SUCCESS;
}

int SUT_No_sigchld( )
{
#ifdef SUT_SIGCHLD_HANDLER
(void) signal( SIGCHLD, SIG_IGN );
#endif
myjobs = 0;
}

/****************************************************************************
  Write data 
 ****************************************************************************/

/* Write the contents of the linebuf to the given fd.  Use the data
   transfer protocol */
int SUT_Write( jobs, i, fd )
SUT_Jobs *jobs;
int      i, fd;
{
SUT_Connection *conn = jobs->conn + i;

if (SUT_Debug & SUT_PASSDATA) 
    fprintf( stderr, "[%d] Writing data to node %d (%d bytes) in SUT_Write\n", 
	    SUT_Mynode, conn->nodenum, conn->linepos );
if (conn->linepos) {
    if (fd == 1) 
	(*jobs->writeparent)( fd, conn->linebuf, conn->linepos );
    else
	SUT_Writedata( fd, conn->linebuf, conn->linepos );
    }
conn->linepos = 0;
}

/****************************************************************************
 This routine copies data from a given fd and writes to another given
 fd, adding "%4d ", where the value is the value of the nodenumber.
 Returns the amount of data read
 ****************************************************************************/
int SUT_CopyToOut( jobs, i )
SUT_Jobs *jobs;
int      i;
{
SUT_Connection *conn = jobs->conn + i;
char tmpbuf[6], *p, *ps;
char buf[SUT_MAX_BUF+6];  /* Used to make line-atomic output */
int len, j;

len = read( conn->readfd, conn->linebuf + conn->linepos, 
	    conn->linelen - conn->linepos );
if (len > 0) {
    if (SUT_Debug & SUT_PASSDATA) {
	fprintf( stderr, 
		"[%d] Writing data to node %d (%d bytes) in SUT_CopyToOut\n", 
		 SUT_Mynode, conn->nodenum, len );
	conn->linebuf[conn->linepos + len] = 0;
	fprintf( stderr, "[%d] data = |%s|\n", SUT_Mynode, 
		 conn->linebuf + conn->linepos );
	}
	
    /* Eventually, this is going to need a handshake to 
       manage data */
    j   = 0;
    p   = ps = conn->linebuf + conn->linepos;
    /* Get the node number */
    sprintf( tmpbuf, "%4d ", jobs->mynodenum );
    strcpy( buf, tmpbuf );

    while (j < len) {
#ifdef FOO
	if (conn->saw_newline) {
	    (*jobs->writeparent)( 1, tmpbuf, 5 );
	    conn->saw_newline = 0;
	    }
#endif
	if (*p == '\n') {
	    if (conn->saw_newline && (int)(p-ps) < SUT_MAX_BUF) {
		strncpy( buf+5, ps, 1 + (int)(p - ps) );
		(*jobs->writeparent)( 1, buf, 5 + 1 + (int)(p - ps) );
		}
	    else {
		if (conn->saw_newline) 
		    (*jobs->writeparent)( 1, tmpbuf, 5 );
		(*jobs->writeparent)( 1, ps, 1 + (int)(p - ps) );
		conn->saw_newline = 1;
		}
	    ps			 = p+1;
	    }
	p++;
	j++;
	}
    /* flush any leftover */
    if (p > ps) {
	(*jobs->writeparent)( 1, ps, (int)(p - ps) );
	}
    }
return len;
}	

int SUT_ProcessInput( jobs, i )
SUT_Jobs *jobs;
int      i;
{
SUT_Header buf;
int        len, err;
SUT_Connection *conn = jobs->conn + i;

if (SUT_Debug & SUT_PASSDATA) 
    fprintf( stderr, "[%d] starting process input\n", 
	     SUT_Mynode );
len = read( conn->readfd, &buf, sizeof(SUT_Header) );
if (len <= 0) return len;

if (len != sizeof(SUT_Header)) {
    fprintf( stderr, "Read partial header (only %d bytes); internal error\n",
	     len );
    return 0;
    }

if (!conn->ops) {
    fprintf( stderr, "Connection functions not defined (conn %d)!\n", i );
    exit(1);
    }
if (!conn->ops->msghandler[buf.op]) {
    fprintf( stderr, "Function not defined for %d!\n", buf.op );
    }
(*(conn->ops->msghandler[buf.op]))( jobs, i, &buf, (void *)0 );
return len;
}

/****************************************************************************
 SUT versions of systems calls to simplify debugging.  They may
 use the value of SUT_Debug (set by SUT_Getdebug) to control
 their behavior.
 ****************************************************************************/
int SUT_Test_env( name )
char *name;
{
extern char *getenv();
char *p = getenv( name );

DEBUG(printf("Value of %s is %s.\n", name, p ? p : "<null>" );)
if (!p || strcmp( p, "NO" ) == 0 || strcmp( p, "no" ) == 0) return 0;
return 1;
}

int SUT_Debug = 0;
int SUT_Mynode = -1;
int SUT_Get_debug()
{
if (SUT_Test_env( "SUT_NOEXEC" ))     SUT_Debug |= SUT_EXECVP;
if (SUT_Test_env( "SUT_FORK" ))       SUT_Debug |= SUT_FORK;
if (SUT_Test_env( "SUT_PRINT" ))      SUT_Debug |= SUT_PRINT;
if (SUT_Test_env( "SUT_HANDSHAKE" ))  SUT_Debug |= SUT_HANDSHAKE;
if (SUT_Test_env( "SUT_PASSDATA" ))   SUT_Debug |= SUT_PASSDATA;
if (SUT_Test_env( "SUT_EXITOPS" ))    SUT_Debug |= SUT_EXITOPS;

if (SUT_Debug) 
    SUT_Mynode = SUT_Mynodenum();
}

/****************************************************************************
 SUT's behavior may be modified by various parameters
 ****************************************************************************/
int SUT_Options = 0;
#define SUT_REGULATE   0x1
int SUT_Get_options()
{
if (SUT_Test_env( "SUT_REGULATE" ))   SUT_Options |= SUT_REGULATE;
}


/* Version fork with some output */
int SUT_fork()
{
if (SUT_Debug & SUT_PRINT) {
    fprintf( stderr, "[%d] Forking...\n", SUT_Mynode );
    if (SUT_Debug & SUT_FORK)
	return 0;
    }
if (SUT_Debug & SUT_FORK) {
    fprintf( stderr, "[%d] Forking...\n", SUT_Mynode );
    return 0;
    }
return fork();
}

/* Do a fork, and return pid, and for parent,  read, and  write fds */
int SUT_Fork_with_pipes( pid, rd, wt )
int *pid, *rd, *wt;
{
int read_pipe[2], write_pipe[2];

if((pipe(read_pipe) == -1) || (pipe(write_pipe) == -1))
    return SUT_PIPE_FAILED;

/* Create a new job */
*pid = SUT_fork();
switch (*pid) {
    case 0:  /* Child */
    close(read_pipe[0]);     /* close unused portion of pipe */
    close(write_pipe[1]);    /* close unused portion of pipe */
    close(1);                /*         close stdout         */
    dup2(read_pipe[1],1);    /*  read_pipe[1] is now stdout  */
    close(read_pipe[1]);     /*  now only duplicate remains  */
    close(0);                /*         close  stdin         */
    dup2(write_pipe[0],0);   /*  write_pipe[0] is now stdin  */
    close(write_pipe[0]);    /*  now only duplicate remains  */
    return SUT_SUCCESS;
	   
    case -1:
    return SUT_FORK_FAILED;
    
    default:
    close(read_pipe[1]);  /* close unused portion of pipe */
    close(write_pipe[0]); /* close unused portion of pipe */
    *rd = read_pipe[0];
    *wt = write_pipe[1];
    } /* End of switch on select for fork return */
return SUT_SUCCESS;
}

/* This allows us to debug by dynamically replacing execvp by a
   dummy routine */
int SUT_execvp( pgm, args )
char *pgm, **args;
{
char **p;
if (SUT_Debug & (SUT_PRINT | SUT_EXECVP)) {
    p = args;
    fprintf( stderr, "[%d] execvp %s ", SUT_Mynode, pgm );
    while (*p) fprintf( stderr, "%s ", *p++ );
    fprintf( stderr, "\n" );
    if (SUT_Debug & SUT_EXECVP)
	return 0;
    }
return execvp( pgm, args );
}


/****************************************************************************
 Error handling
 ****************************************************************************/
int SUT_Errormessage( fd, err )
FILE *fd;
int  err;
{
switch (err) {
    case SUT_SUCCESS:
    fprintf( fd, "No error=n" );
    break;
    case SUT_INVALID_RANGE_SPEC:
    case SUT_RANGES_INVALID:
    fprintf( fd, "Specified node range was invalid\n" );
    break;
    case SUT_NO_MEMORY:
    fprintf( fd, "A malloc failed\n" );
    break;
    case SUT_PIPE_FAILED:
    fprintf( fd, 
	    "Creation of a pipe failed; probably too many files in use\n" );
    break;
    case SUT_FORK_FAILED:
    fprintf( fd, "A fork failed; probably too many processes running\n" );
    break;
    case SUT_SELECT_FAILED:
    fprintf( fd, "A select failed; internal error!\n" );
    break;
    case SUT_GET_EXE_FAILED:
    fprintf( fd, "Unable to generate command line for program to call\n" );
    break;
    case SUT_EXEC_FAILED:
    fprintf( fd, "Unable to exec program.  Program may not exist\n" );
    break;
    case SUT_HANDSHAKE_FAILED:
    fprintf( fd, 
	    "Program did not provide proper handshake or did not start\n" );
    break;
    case SUT_CHILD_UNRESPONSIVE:
    fprintf( fd, "Child on remote node did not respond\n" );
    break;
    case SUT_CHILD_CONFUSED:
    fprintf( fd, "Child on remote node provided wrong handshake\n" );
    break;
    case SUT_PARENT_CONFUSED:
    fprintf( fd, "Parent provided wrong handshake\n" );
    break;

    case SUT_NOT_LOCAL:
    fprintf( fd, "Node not local (not an error)\n" );
    break;
    
    case SUT_CHILD_DONE:
    fprintf( fd, "Child has completed (not an error)\n" );
    break;

    case SUT_UNEXPECTED_EOF:
    fprintf( fd, "Read EOF when expecting data\n" );
    break;
    
    case SUT_READ_FAILED:
    fprintf( fd, "Read failed when expecting data\n" );
    
    default:
    fprintf( fd, "Error %d unknown!\n", err );
    break;
    }
}
